/* DEFINE_MATH_BUILTIN (CODE, NAME, ARGTYPE)
   NAME	  The name of the builtin
   SNAME  The name of the builtin as a string
   ARGTYPE The type of the arguments.  See f95-lang.c

   Use DEFINE_MATH_BUILTIN_C if the complex versions of the builtin are
   also available.  */
DEFINE_MATH_BUILTIN_C (ACOS,  "acos",   0)
DEFINE_MATH_BUILTIN_C (ASIN,  "asin",   0)
DEFINE_MATH_BUILTIN_C (ATAN,  "atan",   0)
DEFINE_MATH_BUILTIN   (ATAN2, "atan2",  1)
DEFINE_MATH_BUILTIN_C (COS,   "cos",    0)
DEFINE_MATH_BUILTIN_C (COSH,  "cosh",   0)
DEFINE_MATH_BUILTIN_C (EXP,   "exp",    0)
DEFINE_MATH_BUILTIN_C (LOG,   "log",    0)
DEFINE_MATH_BUILTIN_C (LOG10, "log10",  0)
DEFINE_MATH_BUILTIN_C (SIN,   "sin",    0)
DEFINE_MATH_BUILTIN_C (SINH,  "sinh",   0)
DEFINE_MATH_BUILTIN_C (SQRT,  "sqrt",   0)
DEFINE_MATH_BUILTIN_C (TAN,   "tan",    0)
DEFINE_MATH_BUILTIN_C (TANH,  "tanh",   0)
DEFINE_MATH_BUILTIN   (J0,    "j0",     0)
DEFINE_MATH_BUILTIN   (J1,    "j1",     0)
DEFINE_MATH_BUILTIN   (JN,    "jn",     2)
DEFINE_MATH_BUILTIN   (Y0,    "y0",     0)
DEFINE_MATH_BUILTIN   (Y1,    "y1",     0)
DEFINE_MATH_BUILTIN   (YN,    "yn",     2)
DEFINE_MATH_BUILTIN   (ERF,   "erf",    0)
DEFINE_MATH_BUILTIN   (ERFC,  "erfc",   0)
